@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qv4vme_moth_p.h"
 
@@ -60,6 +24,7 @@
 #include <private/qv4alloca_p.h>
 #include <private/qqmljavascriptexpression_p.h>
 #include <private/qv4qmlcontext_p.h>
+#include <QtQml/private/qv4runtime_p.h>
 #include <iostream>
 
 #if QT_CONFIG(qml_jit)
@@ -70,6 +35,9 @@
 
 #undef COUNT_INSTRUCTIONS
 
+Q_TRACE_POINT(qtqml, QQmlV4_function_call_entry, const QV4::ExecutionEngine *engine, const QString &function, const QString &fileName, int line, int column)
+Q_TRACE_POINT(qtqml, QQmlV4_function_call_exit)
+
 enum { ShowWhenDeoptimiationHappens = 0 };
 
 extern "C" {
@@ -343,7 +311,7 @@ static struct InstrCount {
     }
 #endif
 
-static inline QV4::Value &stackValue(QV4::Value *stack, size_t slot, const CppStackFrame *frame)
+static inline QV4::Value &stackValue(QV4::Value *stack, size_t slot, const JSTypesStackFrame *frame)
 {
     Q_ASSERT(slot < CallData::HeaderSize() / sizeof(QV4::StaticValue)
                     + frame->jsFrame->argc()
@@ -360,7 +328,7 @@ static inline QV4::Value &stackValue(QV4::Value *stack, size_t slot, const CppSt
 #undef CHECK_EXCEPTION
 #endif
 #define CHECK_EXCEPTION \
-    if (engine->hasException || engine->isInterrupted.loadAcquire()) \
+    if (engine->hasException || engine->isInterrupted.loadRelaxed()) \
         goto handleUnwind
 
 static inline Heap::CallContext *getScope(QV4::Value *stack, int level)
@@ -382,20 +350,18 @@ static inline const QV4::Value &constant(Function *function, int index)
 static bool compareEqualInt(QV4::Value &accumulator, QV4::Value lhs, int rhs)
 {
   redo:
-    switch (lhs.quickType()) {
-    case QV4::Value::QT_ManagedOrUndefined:
-        if (lhs.isUndefined())
-            return false;
-        Q_FALLTHROUGH();
-    case QV4::Value::QT_ManagedOrUndefined1:
-    case QV4::Value::QT_ManagedOrUndefined2:
-    case QV4::Value::QT_ManagedOrUndefined3:
+    if (lhs.isUndefined())
+        return false;
+    if (lhs.isManagedOrUndefined()) {
         // LHS: Managed
         if (lhs.m()->internalClass->vtable->isString)
             return RuntimeHelpers::stringToNumber(static_cast<String &>(lhs).toQString()) == rhs;
         accumulator = lhs;
         lhs = QV4::Value::fromReturnedValue(RuntimeHelpers::objectDefaultValue(&static_cast<QV4::Object &>(accumulator), PREFERREDTYPE_HINT));
         goto redo;
+    }
+
+    switch (lhs.quickType()) {
     case QV4::Value::QT_Empty:
         Q_UNREACHABLE();
     case QV4::Value::QT_Null:
@@ -439,18 +405,18 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
     ExecutionEngineCallDepthRecorder executionEngineCallDepthRecorder(engine);
 
     Function *function = frame->v4Function;
-    Q_ASSERT(function->aotFunction);
+    Q_ASSERT(function->typedFunction);
     Q_TRACE_SCOPE(QQmlV4_function_call, engine, function->name()->toQString(),
                   function->executableCompilationUnit()->fileName(),
                   function->compiledFunction->location.line(),
                   function->compiledFunction->location.column());
     Profiling::FunctionCallProfiler profiler(engine, function); // start execution profiling
 
-    const qsizetype numFunctionArguments = function->aotFunction->argumentTypes.size();
+    const qsizetype numFunctionArguments = function->typedFunction->argumentTypes.size();
 
     Q_ALLOCA_DECLARE(void *, transformedArguments);
     for (qsizetype i = 0; i < numFunctionArguments; ++i) {
-        const QMetaType argumentType = function->aotFunction->argumentTypes[i];
+        const QMetaType argumentType = function->typedFunction->argumentTypes[i];
         if (frame->argc() > i && argumentType == frame->argTypes()[i])
             continue;
 
@@ -459,7 +425,11 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
             memcpy(transformedArguments, frame->argv(), frame->argc() * sizeof(void *));
         }
 
-        Q_ASSERT(argumentType.sizeOf() > 0);
+        if (argumentType.sizeOf() == 0) {
+            transformedArguments[i] = nullptr;
+            continue;
+        }
+
         Q_ALLOCA_VAR(void, arg, argumentType.sizeOf());
 
         if (argumentType == QMetaType::fromType<QVariant>()) {
@@ -467,6 +437,11 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
                 new (arg) QVariant(frame->argTypes()[i], frame->argv()[i]);
             else
                 new (arg) QVariant();
+        } else if (argumentType == QMetaType::fromType<QJSPrimitiveValue>()) {
+            if (frame->argc() > i)
+                new (arg) QJSPrimitiveValue(frame->argTypes()[i], frame->argv()[i]);
+            else
+                new (arg) QJSPrimitiveValue();
         } else {
             argumentType.construct(arg);
             if (frame->argc() > i)
@@ -476,7 +451,7 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
         transformedArguments[i] = arg;
     }
 
-    const QMetaType returnType = function->aotFunction->returnType;
+    const QMetaType returnType = function->typedFunction->returnType;
     const QMetaType frameReturn = frame->returnType();
     Q_ALLOCA_DECLARE(void, transformedResult);
     if (frame->returnValue() && returnType != frameReturn) {
@@ -495,7 +470,8 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
 
     aotContext.engine = engine->jsEngine();
     aotContext.compilationUnit = function->executableCompilationUnit();
-    function->aotFunction->functionPtr(
+
+    function->typedFunction->functionPtr(
                 &aotContext, transformedResult ? transformedResult : frame->returnValue(),
                 transformedArguments ? transformedArguments : frame->argv());
 
@@ -509,6 +485,17 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
                     = (resultObj && resultObj->metaObject()->inherits(frameReturn.metaObject()))
                             ? resultObj
                             : nullptr;
+        } else if (returnType == QMetaType::fromType<QVariant>()) {
+            const QVariant *resultVariant = static_cast<QVariant *>(transformedResult);
+            if (resultVariant->metaType() == frameReturn) {
+                frameReturn.construct(frame->returnValue(), resultVariant->data());
+            } else {
+                // Convert needs a pre-constructed target.
+                frameReturn.construct(frame->returnValue());
+                QMetaType::convert(resultVariant->metaType(), resultVariant->data(),
+                               frameReturn, frame->returnValue());
+            }
+            resultVariant->~QVariant();
         } else {
             // Convert needs a pre-constructed target.
             frameReturn.construct(frame->returnValue());
@@ -519,8 +506,11 @@ void VME::exec(MetaTypesStackFrame *frame, ExecutionEngine *engine)
 
     if (transformedArguments) {
         for (int i = 0; i < numFunctionArguments; ++i) {
-            if (i >= frame->argc() || transformedArguments[i] != frame->argv()[i])
-                function->aotFunction->argumentTypes[i].destruct(transformedArguments[i]);
+            void *arg = transformedArguments[i];
+            if (arg == nullptr)
+                continue;
+            if (i >= frame->argc() || arg != frame->argv()[i])
+                function->typedFunction->argumentTypes[i].destruct(arg);
         }
     }
 }
@@ -557,7 +547,7 @@ ReturnedValue VME::exec(JSTypesStackFrame *frame, ExecutionEngine *engine)
         debugger->enteringFunction();
 
     ReturnedValue result;
-    Q_ASSERT(!function->aotFunction);
+    Q_ASSERT(function->kind != Function::AotCompiled);
     if (function->jittedCode != nullptr && debugger == nullptr) {
         result = function->jittedCode(frame, engine);
     } else {
@@ -884,12 +874,6 @@ QV4::ReturnedValue VME::interpret(JSTypesStackFrame *frame, ExecutionEngine *eng
         CHECK_EXCEPTION;
     MOTH_END_INSTR(CallPropertyLookup)
 
-    MOTH_BEGIN_INSTR(CallElement)
-        STORE_IP();
-        acc = Runtime::CallElement::call(engine, STACK_VALUE(base), STACK_VALUE(index), stack + argv, argc);
-        CHECK_EXCEPTION;
-    MOTH_END_INSTR(CallElement)
-
     MOTH_BEGIN_INSTR(CallName)
         STORE_IP();
         acc = Runtime::CallName::call(engine, name, stack + argv, argc);
@@ -1395,10 +1379,7 @@ QV4::ReturnedValue VME::interpret(JSTypesStackFrame *frame, ExecutionEngine *eng
         const Value left = STACK_VALUE(lhs);
         double base = left.toNumber();
         double exp = ACC.toNumber();
-        if (qIsInf(exp) && (base == 1 || base == -1))
-            acc = Encode(qQNaN());
-        else
-            acc = Encode(pow(base,exp));
+        acc = Encode(QQmlPrivate::jsExponentiate(base, exp));
     MOTH_END_INSTR(Exp)
 
     MOTH_BEGIN_INSTR(Mul)
@@ -1524,7 +1505,7 @@ QV4::ReturnedValue VME::interpret(JSTypesStackFrame *frame, ExecutionEngine *eng
         // We do start the exception handler in case of isInterrupted. The exception handler will
         // immediately abort, due to the same isInterrupted. We don't skip the exception handler
         // because the current behavior is easier to implement in the JIT.
-        Q_ASSERT(engine->hasException || engine->isInterrupted.loadAcquire() || frame->unwindLevel);
+        Q_ASSERT(engine->hasException || engine->isInterrupted.loadRelaxed() || frame->unwindLevel);
         if (!frame->unwindHandler) {
             acc = Encode::undefined();
             return acc;
